package client;

import java.io.IOException;
import java.rmi.NotBoundException;
import java.rmi.registry.LocateRegistry;
import java.rmi.registry.Registry;
import java.util.Scanner;

import transaction.IResourceManager;
import transaction.InvalidTransactionException;
import transaction.ShutdownException;
import transaction.TransactionAbortedException;

/**
 * A console to operate on a <b>remote</b> resource manager. Use this
 * console to simulate single threaded access to a
 * resource manager. To connect to a remote RM use the conn command. See below.
 * The console accepts the following
 * command syntax:
 * 
 * <ul>
 * <li>quit console: q
 * <li>connect to remote RM: conn <RMI_NAME>
 * <li>add resource: ar <rid> <num> <price>
 * <li>update resource: ur <rid> <num> <price>
 * <li>add customer: nc <custName>
 * <li>reserve resource: res <rid> <custName>
 * <li>query bill: bill <custName> 
 * <li>query availability: av <rid>
 * <li>query price: pr <rid>
 * <li>commit
 * <li>abort
 * <li>dienow
 * <li>diebefore
 * <li>dieafter
 * </ul>
 * @author bross
 *
 */
public class ResourceManagerRemoteConsole {
	
	IResourceManager rm = null;
	
	private boolean connected(){
		if(rm == null){
			System.out.println("not connected to any resource manager, please call conn <RMI_NAME>");
			return false;
		}
		
		return true;
	}
	
	private void run(){
		String command;
		String[] splitted;
		Scanner in = new Scanner(System.in);
		int currentXID = 0;
		while(true){			
			command = in.nextLine();
			
			if(command.length() == 0)
				continue;
			
			splitted = command.split("\\s");
			try {
				if(rm != null && currentXID == 0 && (
						splitted[0].equals("ar") || splitted[0].equals("ur") || splitted[0].equals("nc")
						|| splitted[0].equals("res") || splitted[0].equals("bill") || splitted[0].equals("av")
						|| splitted[0].equals("pr"))){
					currentXID = rm.begin();
					System.out.println("started new transaction with xid " + currentXID);
				}
				
				if(splitted[0].equals("q")){										
					if(currentXID != 0 && rm != null){
						rm.abort(currentXID);
						System.out.println("current transaction aborted");
					}
					return;
				}
				
				else if(splitted[0].equals("shutdown")){
					
					if(!connected())
						continue;
					
					System.out.println("shutting down resource manager...");
					if(currentXID != 0){
						rm.abort(currentXID);
						System.out.println("current transaction aborted");
					}
					
					rm.shutdown();
					System.out.println("rm is shut down");
					rm.dieNow();	
					
					rm = null;
					
					return;
				}
				
				else if(splitted[0].equals("conn")){
					if(splitted.length != 2){
						System.err.println("wrong number of arguments, syntax is: conn RMI_NAME");
						continue;
					}
					
					//need to abort current transaction first
					if(rm != null && currentXID != 0){
						System.out.println("aborting current transaction");
						rm.abort(currentXID);
						System.out.println("current transaction aborted");
						currentXID = 0;
					}
					
					try {		
						String name = splitted[1];
						if(!name.equals(IResourceManager.RMI_NAME_CAR_RM) && 
								!name.equals(IResourceManager.RMI_NAME_FLIGHT_RM) &&
								!name.equals(IResourceManager.RMI_NAME_HOTEL_RM)){
							System.out.println("unknown resource manager: " + name);
							System.out.println("valid names are: " + IResourceManager.RMI_NAME_HOTEL_RM + ", " +
									IResourceManager.RMI_NAME_FLIGHT_RM + ", " + IResourceManager.RMI_NAME_CAR_RM);
							continue;
						}
						Registry registry = LocateRegistry.getRegistry();
						rm = (IResourceManager) registry.lookup(name);
						System.out.println("successfully connected");
					} catch (IOException e) {
						throw new RuntimeException(e);
					} catch (NotBoundException e) {
						// TODO Auto-generated catch block
						e.printStackTrace();
					}
					
				}
				else if(splitted[0].equals("ar")){
					if(!connected())
						continue;
					
					if(splitted.length != 4){
						System.err.println("wrong number of arguments, syntax is: ar rid num price");
						continue;
					}
					
					boolean success = rm.addResource(currentXID, splitted[1],
							Integer.parseInt(splitted[2]),
							Integer.parseInt(splitted[3]));
					
					if(success)
						System.out.println("added resource successfully");
					else
						System.out.println("resource already exists");
				}
				else if(splitted[0].equals("ur")){
					if(!connected())
						continue;
					
					if(splitted.length != 4){
						System.err.println("wrong number of arguments, syntax is: ur rid num price");
						continue;
					}
					
					boolean success = rm.updateResource(currentXID, splitted[1],
							Integer.parseInt(splitted[2]),
							Integer.parseInt(splitted[3]));
					
					if(success)
						System.out.println("updated resource successfully");
					else
						System.out.println("update unsuccessful, probably too many reservations...");
				}
				else if(splitted[0].equals("nc")){
					if(!connected())
						continue;
					
					if(splitted.length != 2){
						System.err.println("wrong number of arguments, syntax is: nc custName");
						continue;
					}
					
					boolean success = rm.newCustomer(currentXID, splitted[1]);
					
					if(success)
						System.out.println("added customer successfully");
					else
						System.out.println("insert unsuccessfull, maybe customer exists already");
				}
				else if(splitted[0].equals("res")){
					if(!connected())
						continue;
					
					if(splitted.length != 3){
						System.err.println("wrong number of arguments, syntax is: res rid custName");
						continue;
					}
					
					boolean success = rm.reserveResource(currentXID, splitted[1], splitted[2]);
					if(success)
						System.out.println("reservation successfull");
					else
						System.out.println("reservation unsuccessfull, maybe customer or resource do not exist or resource not available");
				}
				else if(splitted[0].equals("bill")){
					if(!connected())
						continue;
					
					if(splitted.length != 2){
						System.err.println("wrong number of arguments, syntax is: res custName");
						continue;
					}
					
					int bill = rm.queryCustomerBill(currentXID, splitted[1]);
					if(bill >= 0)
						System.out.println("current bill is " + bill);
					else
						System.out.println("unknown customer");
				}
				else if(splitted[0].equals("av")){
					if(!connected())
						continue;
					
					if(splitted.length != 2){
						System.err.println("wrong number of arguments, syntax is: av rid");
						continue;
					}
					
					int av = rm.queryResourceAvailability(currentXID, splitted[1]);
					if(av >= 0)
						System.out.println("current availability is: " + av);
					else
						System.out.println("unknown resource");
				}
				else if(splitted[0].equals("pr")){
					if(!connected())
						continue;
					
					if(splitted.length != 2){
						System.err.println("wrong number of arguments, syntax is: pr rid");
						continue;
					}
					
					int price = rm.queryResourcePrice(currentXID, splitted[1]);
					if(price >= 0)
						System.out.println("current price is: " + price);
					else
						System.out.println("unknown resource");
				}
				else if(splitted[0].equals("commit")){
					if(!connected())
						continue;
					
					if(currentXID != 0){
						boolean c = rm.commit(currentXID);
						if(c){
							System.out.println("transaction successfully committed");
							currentXID = 0;
						}
						else
							System.out.println("transaction NOT committed");
					}
					
				}
				else if(splitted[0].equals("abort")){
					if(!connected())
						continue;
					
					if(currentXID != 0){
						rm.abort(currentXID);
						currentXID = 0;
						System.out.println("transaction successfully aborted");					
					}
				}
				else if(splitted[0].equals("dienow")){
					if(!connected())
						continue;
					
					rm.dieNow();
				}
				else if(splitted[0].equals("diebefore")){
					if(!connected())
						continue;
					
					rm.dieBeforePointerSwitch();
				}
				else if(splitted[0].equals("dieafter")){
					if(!connected())
						continue;
					
					rm.dieAfterPointerSwitch();
				}
				else{
					System.err.println("unknown command");
				}
			} catch (NumberFormatException e) {
				System.err.println("invalid argument: " + e.getMessage());
			} catch (ShutdownException e) {
				System.err.println("resource manager is shutting down");
			} catch (TransactionAbortedException e) {
				currentXID = 0;
				System.err.println("transaction has been aborted by rm");
			} catch (InvalidTransactionException e) {
				//must not happen
				throw new RuntimeException(e);
//			} catch (RemoteException e) {
//				System.out.println("error calling resource manager. maybe it is down...");
			}
		}
	}
	
	/**
	 * @param args
	 */
	public static void main(String[] args) {
		
		ResourceManagerRemoteConsole rc = new ResourceManagerRemoteConsole();
		rc.run();
	}

}
