package transaction;

import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.rmi.RemoteException;
import java.util.HashSet;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.Set;
import java.util.concurrent.atomic.AtomicInteger;

import lockmgr.DeadlockException;
import lockmgr.LockManager;
import lockmgr.LockType;
import transaction.LogEntry.LogType;

public class ResourceManagerImpl implements IResourceManager {
	
	protected Hashtable<Integer, TransactionStatus> transactions =
		new Hashtable<Integer, TransactionStatus>();
	
	protected final AtomicInteger xidGenerator = new AtomicInteger(0);
	
	protected Hashtable<String, ResourceData> resources = 
		new Hashtable<String, ResourceData>();
	
	protected Hashtable<String, Set<String>> reservations = 
		new Hashtable<String, Set<String>>(); 
	
	protected final LockManager lockManager = new LockManager();
	
	protected final LogManager logManager = new LogManager();
	
	protected volatile boolean shutdown = false;
	
	protected int fileSwitch = 0;
	
	protected File rmState;
	protected File db0;
	protected File db1;
	
	protected boolean dieBefore = false;
	protected boolean dieAfter = false;
	
	public ResourceManagerImpl(){
	}

	public ResourceManagerImpl(String db0, String db1, String rmState)
	throws IOException {		 
		 this.rmState = new File(rmState);
		 this.db0 = new File(db0);
		 this.db1 = new File(db1);	
		 
		 readState();
		 readDatabase();
	}
	
	@Override
	public void start(String workingdir) {		
		if(!workingdir.endsWith("/"))
			workingdir += "/";
		
		 this.rmState = new File(workingdir + "rmState");
		 this.db0 = new File(workingdir + "db0");
		 this.db1 = new File(workingdir + "db1");	
		 
		 try {
			 readState();
			 readDatabase();
		} catch (IOException e) {
			throw new RuntimeException(e);
		}
				
	}

	@Override
	public void abort(int xid) throws InvalidTransactionException {
				
		TransactionStatus status = transactions.get(xid);
    	if(status == null)
    		throw new InvalidTransactionException(xid, "unknown transaction");    	    	    	
    	
    	synchronized (transactions) {
    		transactions.put(xid, TransactionStatus.aborted);
    		transactions.notifyAll();
		}
    	
		logManager.deleteLogsForTransaction(xid);
		lockManager.unlockAll(xid);

	}


	@Override
	public int begin() throws ShutdownException{
		
		if(shutdown)
			throw new ShutdownException();
		
		int xid = xidGenerator.incrementAndGet();
		transactions.put(xid, TransactionStatus.running);
		return xid;
	}

	@Override
	public boolean commit(int xid) throws TransactionAbortedException,
			InvalidTransactionException {
		
		validityCheck(xid);
				
		synchronized(resources){
			updateDatabase(xid);
			try {
				writeShadow();
				if(dieBefore)
					System.exit(-1);
				switchPointer();
				if(dieAfter)
					System.exit(-1);
			} catch (IOException e) {
				e.printStackTrace();
								
				throw new TransactionAbortedException(xid, "error while updating db on disk");
			}			
			finally{
				logManager.deleteLogsForTransaction(xid);
				lockManager.unlockAll(xid);
				synchronized (transactions) {
					transactions.remove(xid);
					transactions.notifyAll();
				}
				
			}
		}
		
		return true;
	}

	@Override
	public boolean shutdown() {
		shutdown = true;
		
		synchronized (transactions) {
			while(runningTransactions()){
				try {
					System.out.println("still need to wait for running transactions before shutting down...");
					transactions.wait();
				} catch (InterruptedException e) {
					e.printStackTrace();
					return false;
				}
			}
		}
		
		return true;
	}
	
    private boolean runningTransactions() {		
		synchronized (transactions) {
			for (TransactionStatus ts : transactions.values()) {
				if(ts == TransactionStatus.running){
					return true;
				}
			}
		}

		return false;
	}


	/** 
     * Add a new customer to database.  Should return success if
     * customer already exists.
     * 
     * @param xid id of transaction.
     * @param custName name of customer.
     * @return true on success, false on failure.
     *
     * @throws RemoteException on communications failure.
     * @throws TransactionAbortedException if transaction was aborted.
     * @throws InvalidTransactionException if transaction id is invalid.
     */
    public boolean newCustomer(int xid, String custName) 
	throws TransactionAbortedException,
	       InvalidTransactionException{
    	
    	validityCheck(xid);
    	try {
			lockManager.lock(xid, custName, LockType.WRITE);
			if (!logManager.insertedCustomer(xid, custName)) {
				logManager.log(xid, custName, LogType.INSERT_CUSTOMER, null,
						null);
				return true;
			}
			else
				return false;
		} catch (DeadlockException e) {
			abort(xid);
			throw new TransactionAbortedException(xid, "abort due to deadlock");
		}
    }

    /* Return the total price of all reservations held for a customer. */
    public int queryCustomerBill(int xid, String custName)
	    throws TransactionAbortedException,
		   InvalidTransactionException{
    	
    	validityCheck(xid);
    	
    	int bill = 0;
    	ResourceData data = null;    	    	
    	try {
			lockManager.lock(xid, custName, LockType.READ);
			Set<String> customerReservations = getLatestReservations(xid, custName);
			
			if(customerReservations.size() == 0 && !reservations.containsKey(custName))
				return -1;
			
			for (String reservation : customerReservations) {
				lockManager.lock(xid, reservation, LockType.READ);
				data = getLatestResourceData(xid, reservation);
				if(data != null)
					bill += data.price;
			}
		
			
			
			return bill;
		} catch (DeadlockException e) {
			abort(xid);
			throw new TransactionAbortedException(xid, "abort due to deadlock");
		}

    }
    
    public boolean addResource(int xid, String rid, int num, int price)
    throws TransactionAbortedException, InvalidTransactionException{
    	
    	if(num < 1 || price < 0)
    		return false;
    	
    	validityCheck(xid);
    	
    	try {
			lockManager.lock(xid, rid, LockType.WRITE);
			ResourceData data = resources.get(rid);
			if(data == null && logManager.insertedResource(xid, rid) == null){
				data = new ResourceData(num,0,price);
				logManager.log(xid, rid, LogType.INSERT_RESOURCE, data, null);			
				return true;
			}
			else{
				return false;
			}
		} catch (DeadlockException e) {
			abort(xid);
			throw new TransactionAbortedException(xid, "abort due to deadlock");
		}
    }
    
    public boolean updateResource(int xid, String rid, int num, int price)
    throws TransactionAbortedException, InvalidTransactionException{    	    	
    	
    	validityCheck(xid);
    	
    	try {
			lockManager.lock(xid, rid, LockType.WRITE);
			ResourceData data = getLatestResourceData(xid, rid);
			if(data == null)
				return false;
			
			if(-num > data.totalAvailable - data.reserved)
				return false;
			
			logManager.log(xid, rid, LogType.UPDATE_RESOURCE,
					new ResourceData(data.totalAvailable + num, data.reserved, price),null);
						
			return true;
		} catch (DeadlockException e) {
			abort(xid);
			throw new TransactionAbortedException(xid, "abort due to deadlock");
		}				
    }        
    
    public int queryResourceAvailability(int xid, String rid)
    throws TransactionAbortedException, InvalidTransactionException{

    	validityCheck(xid);
    	
    	try {
			lockManager.lock(xid, rid, LockType.READ);
			
			ResourceData data = getLatestResourceData(xid, rid);
						
			if(data == null)
				return -1;
			
			return data.totalAvailable - data.reserved;
		} catch (DeadlockException e) {
			abort(xid);
			throw new TransactionAbortedException(xid, "abort due to deadlock");
		}
    }
    
    public int queryResourcePrice(int xid, String rid)
    throws TransactionAbortedException, InvalidTransactionException{

    	validityCheck(xid);
    	
    	try {
			lockManager.lock(xid, rid, LockType.READ);
			
			ResourceData data = getLatestResourceData(xid, rid);
			
			if(data == null)
				return -1;
			
			return data.price;
		} catch (DeadlockException e) {
			abort(xid);
			throw new TransactionAbortedException(xid, "abort due to deadlock");
		}
    }
    
    public boolean reserveResource(int xid, String rid, String custName)
    throws TransactionAbortedException, InvalidTransactionException{
    	
    	if(rid == null)
			return false;
    	
    	validityCheck(xid);
    	
    	try {
			lockManager.lock(xid, rid, LockType.WRITE);
						
			ResourceData data = getLatestResourceData(xid, rid);			
			
			if(data == null || data.totalAvailable - data.reserved == 0)
				return false;									
			
			lockManager.lock(xid, custName, LockType.WRITE);
			Set<String> customerReservations = reservations.get(custName);
			if(customerReservations == null && !logManager.insertedCustomer(xid, custName))
				return false;
			
			if(getLatestReservations(xid, custName).contains(rid))
				return false;
			
			logManager.log(xid, rid, LogType.UPDATE_RESOURCE,
					new ResourceData(data.totalAvailable, data.reserved + 1, data.price), null);
			logManager.log(xid, custName, LogType.INSERT_RESERVATION, null, rid);
			
			return true;
		} catch (DeadlockException e) {
			abort(xid);
			throw new TransactionAbortedException(xid, "abort due to deadlock");
		}    	
    }
    
    public void dieNow(){
    	System.exit(1);
    }
    
    public void dieBeforePointerSwitch(){
    	dieBefore = true;
    }
    
    public void dieAfterPointerSwitch(){
    	dieAfter = true;
    }
    
    private void validityCheck(int xid) throws TransactionAbortedException, InvalidTransactionException{
    	TransactionStatus status = transactions.get(xid);
    	if(status == null)
    		throw new InvalidTransactionException(xid, "unknown transaction");
    	if(status == TransactionStatus.aborted)
    		throw new TransactionAbortedException(xid, "transaction already aborted");
    }
    
    private void updateDatabase(int xid) {
		LinkedList<LogEntry> redoOperations = logManager.getLogEntries(xid);
		if(redoOperations == null)
			return;
		
		LogEntry current = null;
		Iterator<LogEntry> it = redoOperations.iterator();
		while(it.hasNext()){
			current = it.next();
			switch(current.type){
			case INSERT_CUSTOMER:
				Set<String> customerReservations = reservations.get(current.resourceID);
				if(customerReservations == null)
					reservations.put(current.resourceID, new HashSet<String>());
				break;
			case INSERT_RESOURCE:
				resources.put(current.resourceID, current.afterImage);
				break;			
			case UPDATE_RESOURCE:
				resources.put(current.resourceID, current.afterImage);
				break;
			case INSERT_RESERVATION:
				reservations.get(current.resourceID).add(current.reservation);
				break;			
			}
		}
	}
    
    private ResourceData getLatestResourceData(int xid, String rid){
    	ResourceData data = logManager.insertedResource(xid, rid);
		ResourceData dataUpdate = logManager.getLatestUpdate(xid, rid);
		if(dataUpdate != null)
			data = dataUpdate;
		if(data == null)
			data = resources.get(rid);
		
		return data;
    }
    
    private Set<String> getLatestReservations(int xid, String customer){
    	Set<String> latestReservations = new HashSet<String>();
    	if(reservations.containsKey(customer))
    		latestReservations.addAll(reservations.get(customer));
    	latestReservations.addAll(logManager.getReservations(xid, customer));
    	
    	return latestReservations;
    }
    
    private void readState() throws IOException{
    	if(rmState.exists()){
    		DataInputStream dis = new DataInputStream(new FileInputStream(this.rmState));
    		fileSwitch = dis.readInt();
    		dis.close();
    	}    	
    }
    
    private void writeState() throws IOException{
    	DataOutputStream dos = new DataOutputStream(new FileOutputStream(rmState));
    	dos.writeInt(fileSwitch);
    	dos.close();
    }
    
    private void writeShadow() throws IOException{
    	File db = fileSwitch == 0 ? db1 : db0;
    	ObjectOutputStream oos = new ObjectOutputStream(new FileOutputStream(db));
    	oos.writeObject(resources);
    	oos.writeObject(reservations);
    	oos.close();
    }
    
    private void switchPointer() throws IOException{
    	fileSwitch = fileSwitch == 0 ? 1 : 0;
    	writeState();
    }
    
    @SuppressWarnings("unchecked")
	private void readDatabase() throws IOException{
    	File db = fileSwitch == 0 ? db0 : db1;
    	if(db.exists()){
	    	ObjectInputStream ois = new ObjectInputStream(new FileInputStream(db));
	    	try {
				resources = (Hashtable<String, ResourceData>)ois.readObject();
				reservations = (Hashtable<String, Set<String>>) ois.readObject();
			} catch (ClassNotFoundException e) {
				throw new IOException(e);
			}
	    	ois.close();
    	}
    	else{
    		System.out.println("database file not found, creating new database");
    	}
    }

}
